
import * as fs from 'fs';
import * as path from 'path';

export enum vectorType {
  value,
  function,
}
export interface IRQvector {
  index: number;
  name: string;
  type: vectorType;
  value?: number;
}

export class Pin {
  in?: boolean = false;
  out?: boolean = false;
  pullup?: boolean = false;
  pulldown?: boolean = false;
  irq?: boolean = false;
  adc?: boolean = false;
  dac?: boolean = false;
  wakeup?: boolean = false;
  opendrain?: boolean = false;
}

export class Cpu {

  name: string = "";
  cortex_type: string = "cortex-m23";
  cortex_fpu: string = "softvfp";
  hw: any = {};
  files: { [id: string]: string[] } = {};
  funcs: { [name: string]: string[] } = {};
  code: string[] = [];
  pins: { [id: string]: Pin } = {};
  maps: { [id: string]: { [id: number]: { pin: string } } } = {};
  IRQvectors: IRQvector[] = [];

  constructor(
    name: string,
    IRQvectors: IRQvector[] = [],
    pins: { [id: string]: Pin; } = {},
    maps: { [id: string]: { [id: number]: { pin: string } } } = {},
  ) {
    this.name = name;
    this.pins = pins;
    this.maps = maps;
    this.IRQvectors = IRQvectors;
  }

  add_hw(devs: any) {
    this.hw = {
      ...this.hw,
      ...devs,
    };
  }

  init() {

  }

  gpio_alloc(name: string, pin: string, req: any) {

  }

  set_handler(key: string, value: number) {
    var v = Object.keys(this.IRQvectors).findIndex(k => this.IRQvectors[parseInt(k)].name == key);
    console.log("hit:", key, this.IRQvectors[v]);
  }

  push_function(name: string, s: string) {
    if (!(name in this.funcs))
      this.funcs[name] = [];
    var rows = s.split('\n');
    var snip = 0;
    for (var r of rows) {
      var trimmed = r.trimLeft();
      if (trimmed.length < r.length) {
        snip = r.length - trimmed.length;
        break;
      }
    }
    console.log("snip",snip);

    // for (var r of rows)
    //   this.funcs[name].push(r);
    var cnt = 0;
    for (var r of rows) {
      var s = r.substr(snip);
      if (cnt == 0 && s.length == 0)
        continue;
      this.funcs[name].push(s);
      cnt += 1;
    }
  }

  push_code(fn: string, s: string) {
    if (!(fn in this.files))
      this.files[fn] = [];
    for (var r of s.split('\n'))
      this.files[fn].push(`   ${r.trimLeft()}`)
  }

  push_code_asm(fn: string, s: string) {
    if (!(fn in this.files))
      this.files[fn] = [];
    for (var r of s.split('\n')) {
      if (r.indexOf(':') != -1)
        this.files[fn].push(`${r.trimLeft()}`)
      else
        this.files[fn].push(`   ${r.trimLeft()}`)
    }
  }

  build() {
    this.push_code_asm('vectors.S',`
    .syntax unified
    .cpu ${this.cortex_type}
    .fpu ${this.cortex_fpu}
    .thumb

    .global g_pfnVectors
    .global Default_Handler
    .word _sidata
    .word _sdata
    .word _edata
    .word _sbss
    .word _ebss
    .word _sheap
    .section .text.Reset_Handler
    .weak Reset_Handler
    .type Reset_Handler, %function

    Reset_Handler:
    ldr   r0, =_estack
    mov   sp, r0

    /* Copy the data segment initializers from flash to SRAM */
    ldr r0, =_sdata
    ldr r1, =_edata
    ldr r2, =_sidata
    movs r3, #0
    b LoopCopyDataInit

    CopyDataInit:
    ldr r4, [r2, r3]
    str r4, [r0, r3]
    adds r3, r3, #4

    LoopCopyDataInit:
    adds r4, r0, r3
    cmp r4, r1
    bcc CopyDataInit

    /* Zero fill the bss segment. */
    ldr r2, =_sbss
    ldr r4, =_ebss
    movs r3, #0
    b LoopFillZerobss

    FillZerobss:
    str  r3, [r2]
    adds r2, r2, #4

    LoopFillZerobss:
    cmp r2, r4
    bcc FillZerobss

    /* Call the application's entry point.*/
    bl main

    LoopForever:
    b LoopForever

    .size Reset_Handler, .-Reset_Handler

    .section .text.Default_Handler,"ax",%progbits

    Default_Handler:
    Infinite_Loop:
    b Infinite_Loop
    .size Default_Handler, .-Default_Handler

    .section .isr_vector,"a",%progbits
    .type g_pfnVectors, %object
    .size g_pfnVectors, .-g_pfnVectors


    g_pfnVectors:
    `)
    var nextkey: number = 0;
    for (var key in this.IRQvectors.sort((a, b) => {
      return a.index - b.index;
    })) {
      var vec = this.IRQvectors[key]
      console.log("xx",vec);

      if (nextkey != vec.index) {
        for (var i = nextkey; i < vec.index; i++)
          this.push_code_asm('vectors.S',`.word Default_Handler`)
      }
      if (vec.type == vectorType.function) {
        this.push_code_asm('vectors.S',`.word ${vec.name}_handler`)
      } else {
        this.push_code_asm('vectors.S',`.word ${123}`)
      }
      nextkey = vec.index + 1;
    }
    for (var key in this.IRQvectors) {
      var k: number = parseInt(key);
      var vec = this.IRQvectors[key]
      if (vec.type == vectorType.function) {
        this.push_code_asm('vectors.S',`
          .weak      ${vec.name}_handler
          .thumb_set ${vec.name}_handler,Default_Handler
          `);
      }
    }
  }

  init_build(directory: string) {
    if (!fs.existsSync(directory)) {
      fs.mkdirSync(directory);
    } else {
      var files = fs.readdirSync(directory);
      for (const file of files) {
        //fs.unlinkSync(path.join(directory, file))
        console.log(path.join(directory, file))
      }
    }
  }

  compile(directory: string) {
    this.init_build(`build_${directory}`);

    for (var f in this.files) {
      fs.writeFileSync(`build_${directory}/${f}`, this.files[f].join("\n"));
    }
    var s = "";
    for (var f in this.funcs) {
      s += "\n";
      s += this.funcs[f].join("\n");
      s += "\n";
    }
    fs.writeFileSync(`build_${directory}/main.c`, s);

  }
}
